/*
 * Copyright (C) 1997-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */ 


#include "server.h"

extern struct misc_meta_data *mmd;
extern struct gengetopt_args_info conf;

static int com_db_read(int, int, char **);
static int com_db_write(int, int, char **);
static int com_si(int, int, char **);
static int com_si(int, int, char **);
static int com_version(int, int, char **);
static int com_sb(int, int, char **);
static int com_sc(int, int, char **);
static int com_stat(int, int, char **);
static int com_help(int, int, char **);
static int com_hup(int, int, char **);
static int com_perms(int, int, char **);
static int com_pic(int, int, char **);
static int com_term(int, int, char **);
static int com_cs(int, int, char **);
static int com_streams(int, int, char **);
static int com_ns(int, int, char **);
static int com_ps(int, int, char **);
static int com_play(int, int, char **);
static int com_stop(int, int, char **);
static int com_pause(int, int, char **);
static int com_next(int, int, char **);
static int com_nomore(int, int, char **);
static int com_ff(int, int, char **);
static int com_jmp(int, int, char **);

static struct linked_cmd_list *lcl_head;
static int dbtool_has_pic;

static struct command cmd_struct[] = {
/*
 * List of commands. The first one is just a
 * dummy to let the command nubers start by 1
 */

{
	.name = "dummy"
},

	/* commands that can be handled by the server itself */

	{
		.name = "stat",	
		.handler = com_stat,
		.perms = AFS_READ, 
		.modus = MODUS_READ, 
		.description = "print status info for current song",
		.usage = "stat [n]",
		.help =
                "Without any arguments, stat continuously prints     \n"
                "status messages of the song being streamed. Use the \n"
                "optional number n to let stat exit after having     \n"
                "displayed status n times.                           \n"
	},

	{
		.name = "sb",
		.handler = com_sb,
		.perms = AFS_READ,
		.modus = MODUS_READ, 
		.description = "print status bar for current song",
		.usage = "sb [n]",
		.help =
                "Without any arguments, sb continuously prints a     \n"
                "status bar of the form                              \n"

                        "\n\t\"12:34 [56:12] (56%) filename\"\n\n"

                "indicating playing time, remaining time, percentage \n"
                "and the name of the file beeing streamed.  Use the  \n"
                "optional number n to let stat exit after having     \n"
                "displayed the status bar n times.                   \n"
	},
	{
		.name = "sc",
		.handler = com_sc,
		.perms = AFS_READ,
		.modus = MODUS_READ, 
		.description = "print name of song whenever song changes",
		.usage = "sc",
		.help =
                "sc prints exactly one line (the filename of the song being "
		"played) whenever a song change was detected.\n"
	},

	{	
		.name = "help",	
		.handler = com_help,
		.perms = 0,
		.modus = MODUS_READ, 
		.description = "print help text",
		.usage = "help [command]",
		.help =
                "Without any arguments, help prints a list of        \n"
                "availible commands. When issued with a command name \n"
                "as first argument, help prints out a description for\n"
                "that command.                                       \n"
	},
	{	
		.name = "hup",
		.handler = com_hup,
		.perms = AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "force reload of config file and log file",
		.usage = "hup",
		.help =
                "After rereading the config file, a signal is sent to\n"
                "all children which forces them to close/reopen the  \n"
                "log file.                                           \n"
	},
	{
		.name = "perms",
		.handler = com_perms,
		.perms = 0,
		.modus = MODUS_READ, 
		.description = "show permissions for each command",
		.usage = "perms",
		.help =
                "Print the list of available commands together       \n"
                "with the privileges one must have to execute that   \n"
                "command.                                            \n"
	},
	{
		.name = "streams",
		.handler = com_streams,
		.perms = DB_READ,
		.modus = MODUS_READ, 
		.description = "list streams",
		.usage = "streams",
		.help =
                "Print list of available streams. Use \"cs\" to switch \n"
                "to any of these.                                    \n"
	},


	{
		.name = "term",	
		.handler = com_term,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "terminate server and audio file sender",
		.usage = "term",
		.help =
                "Shuts down the server. Instead of this command, you \n"
                "can also send SIGINT or SIGTERM. It should never be \n"
                "necessary to send SIGKILL.                          \n"
	},
	{
		.name = "si",
		.handler = com_si,
		.perms = 0,
		.modus = MODUS_READ, 
		.description = "print server info",
		.usage = "si",
		.help =
                "Print server uptime and other information.      \n"
	},
	{
		.name = "version",
		.handler = com_version,
		.perms = 0,
		.modus = MODUS_READ, 
		.description = "print server's version",
		.usage = "version",
		.help =	
		"Show version and other info\n"
	},
	{
		.name = "cs",
		.handler = com_cs,
		.perms = AFS_WRITE | DB_READ,
		.modus = MODUS_READ, 
		.description = "change stream",
		.usage = "cs [stream]",
		.help =
                "Selects stream stream or prints current stream when \n"
                "no stream is given.                                 \n"
	},
	{
		.name = "ns",
		.handler = com_ns,
		.perms = AFS_WRITE | DB_READ,
		.modus = MODUS_READ, 
		.description = "change to next stream",
		.usage = "ns",
		.help =
		"Cycle forward through stream list.\n"
	},
	{
		.name = "ps",
		.handler = com_ps,
		.perms = AFS_WRITE | DB_READ,
		.modus = MODUS_READ, 
		.description = "change to previous stream",
		.usage = "ps",
		.help =
		"Cycle backward through stream list.\n"
	},
	{
		.name = "pic",
		.handler = com_pic,
		.perms = DB_READ,
		.modus = MODUS_READ, 
		.description = "show pic of song",
		.usage = "pic [songname]",
		.help =
                "dump jpg image of pic associated with songname to   \n"
                "stdout.                                             \n"
	},

	{
		.name = "jmp",
		.handler = com_jmp,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "jmp to given position in current song",
		.usage = "jmp [n]",
		.help =
		"Jump to position n, where 0 <= n <= 100\n"
	},

	{
		.name = "ff",
		.handler = com_ff,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "jmp amount of time forwards or backwards "
			"in current song",
		.usage = "ff n[-]",
		.help =	
		"Jump n seconds forward or backward.\n"
	},
	{
		.name = "nomore",
		.handler = com_nomore,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "stop playing after current song is over",
		.usage = "nomore", 
		.help = 
		"Use nomore instead of stop if you don't like\n"
		"sudden endings.\n"
	},
	{
		.name = "next",
		.handler = com_next,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "skip rest of current song",
		.usage = "next",
		.help =	
		"If stopped or paused, start playing.\n"
	},
	{
		.name ="pause",
		.handler = com_pause,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "pause current song",
		.usage = "pause",
		.help =	
		"Use play to restart.\n"
	},
	{
		.name = "play",	
		.handler = com_play,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "start playing or resume playing when paused",
		.usage = "play [stream]",
		.help = 
		"Start/continue streaming.\n"
	},
	{
		.name = "stop",	
		.handler = com_stop,
		.perms = AFS_READ | AFS_WRITE,
		.modus = MODUS_READ, 
		.description = "stop playing",
		.usage = "stop",
		.help =	
                "When stopped, server and afs keep running, so you   \n"
                "can still connect to the server and issue commands, \n"
                "eg. play.                                           \n"
	},
	/* this indicates the end of the list. Do not touch. */
	{ 
		.name = "",
		.handler = NULL,
		.perms = 0,
		.description = "",
		.help = ""
	}
};

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mmd_lock();
	*new_mmd = *mmd;
	mmd_unlock();
}

/*
 * construct string containing uptime
 */
static void uptime_str(int i, char *msg)
{
	int n;

	snprintf(msg, MAXLINE - 1, "%i day%s, %i:%02i:%02i%n", i / 86400,
		i > 86400 && i < 2 * 86400?  "" : "s",
		(i / 3600) % 24, (i / 60) % 60, i % 60, &n);
	msg[n] = '\0';
}

/*
 * write human readable string to msg
 */
static void cmd_perms_itohuman(unsigned int perms, char *msg) 
{
	msg[0] = '\0';
	if ((perms & DB_READ)) 
		strcat(msg, "DB_READ ");
	if ((perms & DB_WRITE)) 
		strcat(msg, "DB_WRITE ");
	if ((perms & AFS_READ)) 
		strcat(msg, "AFS_READ ");
	if ((perms & AFS_WRITE)) 
		strcat(msg, "AFS_WRITE ");
	if (!msg[0])
		strcpy(msg, "(none)");
}

/* 
 * compute status bar string 
 */
static void compute_sb_string(const struct misc_meta_data *nmmd, char *dest)
{
	char *name, *base;
	long long unsigned secs, rsecs;

	if (nmmd->chunks_total) {
		secs = (long long) nmmd->seconds_total * nmmd->chunks_sent
		 	/ nmmd->chunks_total;
		rsecs = (long long) nmmd->seconds_total * 
			(nmmd->chunks_total - nmmd->chunks_sent)
			/ nmmd->chunks_total; 
	}
	*dest = '\0';
	if (!(name = strdup(nmmd->filename))) {
		clog(ERROR, "%s: Can't strdup\n", __func__);
		return;
	}
	base = basename(name);
	if (nmmd->chunks_total) {
		sprintf (dest,
			"%llu:%02llu [%llu:%02llu] (%lu%%) %s",
			secs / 60, secs % 60, 
			rsecs / 60, rsecs % 60,
			100 * ((nmmd->chunks_sent + 5) / 10)
				/ ((nmmd->chunks_total + 5) / 10),
			base
		);
	} else
		sprintf(dest, "0:00 [0:00] (100%%) %s", base);
	free(name);
	return;
}

/* compute human readable string containing
 * command handler for given pointer to handler
 */
static char *cmd_handler_itohuman(int (*handler)(int, int, char **)) 
{
	if (handler == com_db_read) 
		return "db_read  ";
	if (handler == com_db_write) 
		return "db_write ";
	return "server   ";

}

static void free_stream_list(char **stream_list)
{
	char **ptr = stream_list;
        clog(DEBUG, "%s: nuking stream list\n", __func__);
        while (*ptr) {
                // clog(DEBUG, "%s: nuking %s\n", __func__, *ptr);
                free(*ptr);
                ptr++;
        }
        free(stream_list);
}

/*
 * get stream list. Must be called with mmd_lock held.
 */
static int get_stream_list(char ***stream_list)
{
	FILE *pipe = NULL;
	char buf[MAXLINE];
	char stream[MAXLINE];
	int num_streams = 0, i;
	char *streams_cmd[] = {"streams", NULL};

	if (!(pipe = popen_read_dbtool(streams_cmd)))
		return -1;
	while (fgets(buf, MAXLINE - 1, pipe) && buf[0] != '#')
		num_streams++;
	pclose(pipe);
	num_streams++;
	if (!(pipe = popen_read_dbtool(streams_cmd)))
		return -1;
	clog(DEBUG, "%s: mallocing mem for %i streamnames\n", __func__,
		num_streams);
	*stream_list = malloc(num_streams * sizeof(char *));
	if (!*stream_list)
		return -1;
	clog(DEBUG, "%s: stream list @%lx, reading streams\n", __func__,
		(long unsigned) stream_list);
	i = 0;
	while (i < num_streams && fgets(buf, MAXLINE - 1, pipe)) {
		buf[MAXLINE - 1] = '\0';
		if (sscanf(buf, "%200s", stream) < 1)
				continue;
		(*stream_list)[i] = strdup(stream);
		//clog(DEBUG, "%s: found stream %s (@%lx)\n", __func__,
		//	(*stream_list)[i], (*stream_list)[i]);
		i++;
	}
	pclose(pipe);
	if (!i)
		clog(CRIT, "%s: no streams found\n", __func__);
	(*stream_list)[i] = NULL;
	return i;
}

static int compute_status_volatile(char *buf, struct misc_meta_data *nmmd)
{
	char status[20]; /* afs status */
	char bar[_POSIX_PATH_MAX + 30];
	char flags[10];
	int n;
	char *tmp = strdup(buf);

	if (!tmp)
		return 0;
	afs_status_tohuman(nmmd->afs_status_flags, status);
	afs_get_status_flags(nmmd->afs_status_flags, flags);
	compute_sb_string(nmmd, bar);
	snprintf(buf, STRINGSIZE - 1,
		"status_bar:%s\n"
		"status:%s\n"
		"status_flags:%s\n"
		"stream:%s\n"
		"format:%s\n"
		"%s"
		"%n", 
		bar,
		status,
		flags,
		nmmd->stream_name,
		audio_format_name(nmmd->audio_format),
		nmmd->audio_file_info,
		&n
	);
	buf[n] = '\0';
	n = strcmp(tmp, buf);
	free(tmp);
	return n;
}

static void compute_status(char *buf, size_t len, struct misc_meta_data *nmmd)
{
	int n;

	snprintf(buf, len - 1,
		"file_size:%lu\n"
		"mtime:%s"
		"length:%i\n"
		"length_min:%i\n"
		"length_sec:%02i\n"
		"num_played:%u\n"
		"score:%d\n%n",
		nmmd->size / 1024,
		nmmd->mtime,
		nmmd->seconds_total,
		nmmd->seconds_total / 60,
		nmmd->seconds_total % 60,
		nmmd->num_played,
		nmmd->score,
		&n
	);
	buf[n] = '\0';
}

static int find_stream(char *name, char **list)
{

	char **p = list;
	int i;

	for (i = 0, p = list; *p; i++, p++)
		if (!strcmp(*p, name))
			return i;
	return -1;
}


/*
 * Send list of permissions for each command to socket.
 */
static int lcl_send_perms(int socket_fd) 
{
        struct linked_cmd_list *lcl_ptr = lcl_head;
        while (lcl_ptr) {
		char buf[STRINGSIZE];
		char perms[MAXLINE];
		int n;

		cmd_perms_itohuman(lcl_ptr->cmd->perms, perms);
		snprintf(buf, STRINGSIZE - 1, "%s\t%s\t%s\t%s\n%n", 
			lcl_ptr->cmd->name, 
			cmd_handler_itohuman(lcl_ptr->cmd->handler), 
			lcl_ptr->cmd->modus == MODUS_READ?  "read" : "write", 
			perms, 
			&n);
		buf[n] = '\0';
		if (send_buffer(socket_fd, buf) < 0)
			return -1;
                lcl_ptr = lcl_ptr->next;
        }
	return 0;
}

/*
 * Send descriptions of all commands to socket.
 */
static int lcl_send_descriptions(int socket_fd) 
{
        struct linked_cmd_list *lcl_ptr = lcl_head;
        while (lcl_ptr) {
		char buf[STRINGSIZE];
		int n;

		snprintf(buf, STRINGSIZE - 1, "%s\t%s\n%n", 
			lcl_ptr->cmd->name, lcl_ptr->cmd->description, &n);
		buf[n] = '\0';
		if (send_buffer(socket_fd, buf) < 0)
			return -1;
                lcl_ptr = lcl_ptr->next;
        }
	return 0;
}

/*
 * Add new entry in linked command list
 */
static void lcl_add(struct command *new)
{
        struct linked_cmd_list *new_lcl, *p;

	clog(DEBUG, "%s: adding %s\n", __func__, new->name);
        new_lcl = malloc(sizeof(struct linked_cmd_list));
        if (!new_lcl) {
		clog(CRIT, "%s: can't add %s\n", __func__, new->name);
                return;
	}
	new_lcl->cmd = new;
	if (!lcl_head) {
		lcl_head = new_lcl;
		new_lcl->next = NULL;
		return;
	}
	if (strcmp(lcl_head->cmd->name, new->name) >= 0) {
		new_lcl->next = lcl_head;
		lcl_head = new_lcl;
		return;
	}
	p = lcl_head;
	while (p->next && strcmp(p->next->cmd->name, new->name) < 0)
		p = p->next;
	new_lcl->next = p->next;
	p->next = new_lcl;
}

/*
 * nuke dbtool entries from list
 */
void lcl_nuke(void)
{
        struct linked_cmd_list *next;
	clog(INFO, "%s: nuking command list\n", __func__);
	while (lcl_head) {
		clog(DEBUG, "%s: nuking %s\n", __func__, lcl_head->cmd->name);
		next = lcl_head->next;
		if (lcl_head->cmd->handler == com_db_read || 
				lcl_head->cmd->handler == com_db_write) {
			free(lcl_head->cmd->name);
			free(lcl_head->cmd->description);
			free(lcl_head->cmd->usage);
			free(lcl_head->cmd->help);
			free(lcl_head->cmd);

			}
		free(lcl_head);
		lcl_head = next;
	}
}

void chop(char* buf)
{
	int n = strlen(buf);
	if (!n)
		return;
	if (buf[n - 1] == '\n')
		buf[n - 1] = '\0';
}

/*
 * Initialize linked command list.
 */
void lcl_init(void) 
{
	int i = 1;
	size_t help_size;
	FILE *pipe = NULL, *pipe2 = NULL;
	char buf[MAXLINE];
	char buf2[MAXLINE];
	char *help_cmd[]={"help", NULL};

	clog(INFO, "%s: init command list\n", __func__);
	/* add internal/afs commands if not already initialized */
	while (cmd_struct[i].name[0]) {
		lcl_add(cmd_struct + i);
		i++;
	}
	/* add dbtool commands */
	dbtool_has_pic = 0;
        if (!(pipe = popen_read_dbtool(help_cmd))) {
		clog(EMERG, "%s: popen_read_dbtool failed\n", __func__); 
		exit(EXIT_FAILURE);
	}
	while ((fgets(buf, MAXLINE - 1, pipe))) {
		struct command *new_cmd = malloc(sizeof(struct command));
		char *help_cmd_cmd[] = {"help", "", NULL};

		new_cmd->description = "";
		new_cmd->usage = "";
		new_cmd->help = "";
		if (sscanf(buf, "%200s %u %u", buf2, &new_cmd->perms,
				&new_cmd->modus) < 3) {
			free(new_cmd);
			continue;
		}
		new_cmd->name = strdup(buf2);
		if (!strcmp(new_cmd->name, "pic")) {
			/* 
			 * omit pic, it is always a valid (internal) 
			 * command 
			 */
			dbtool_has_pic = 1;
			free(new_cmd->name);
			free(new_cmd);
			continue;
		}
		if (new_cmd->modus == MODUS_WRITE)
			new_cmd->handler = com_db_write;
		else
			new_cmd->handler = com_db_read;
		help_cmd_cmd[1] = new_cmd->name;
		if (!(pipe2 = popen_read_dbtool(help_cmd_cmd))) {
			clog(ERROR, "%s: can't read help for %s (%s)\n", 
				__func__, new_cmd->name, strerror(errno)); 
			goto out;
		}
		if (!fgets(buf2 , MAXLINE - 1, pipe2))
			continue;
		chop(buf2);
		new_cmd->description = strdup(buf2);
		if (!fgets(buf2, MAXLINE - 1, pipe2))
			continue;
		chop(buf2);
		new_cmd->usage = strdup(buf2);
		new_cmd->help = NULL;
		help_size = 0;
		while (fgets(buf2, MAXLINE - 1, pipe2)) {
			size_t len = strlen(buf2);
			if (new_cmd->help) {
				help_size = len + strlen(new_cmd->help) + 1;
				new_cmd->help = realloc(new_cmd->help, 
					help_size);
				strcat(new_cmd->help, buf2); 
			} else {
				help_size = len * sizeof(char) + 1;
				new_cmd->help = malloc(help_size);
				strcpy(new_cmd->help, buf2); 
			}
		}
		pclose(pipe2);
		pipe2 = NULL;
		lcl_add(new_cmd);
	}
out:
 	if (pipe)
		pclose(pipe);
 	if (pipe2)
		pclose(pipe2);
}


/*
 * Check if given string belongs is a valid command, i.e. is contained in
 * the list.
 */
static struct linked_cmd_list *lcl_find(char *name) 
{
        struct linked_cmd_list *lcl_ptr = lcl_head;

        while (lcl_ptr && strcmp(lcl_ptr->cmd->name, name)) 
                lcl_ptr = lcl_ptr->next;
        return lcl_ptr;
}
/*
 * dbtool command handler
 */
static int dbtool_cmd_handler(unsigned int modus, int socket_fd, 
		int argc, char **argv) 
{
        char buf[MAXLINE];
	FILE *pipe = NULL;
	int file_no, ret = -1;
	ssize_t n;
	char **new_argv = malloc((argc + 3) * sizeof(char *));
	
	if (!new_argv)
		goto out;
	new_argv[0] = BINDIR "/para_dbtool";
	for (n = 1; n <= argc + 1; n++)
		new_argv[n] = argv[n - 1];
	new_argv[n] = NULL;
	if (modus == MODUS_WRITE)
		pipe = popen_write_vp(new_argv[0], new_argv);
	else
		pipe = popen_read_vp(new_argv[0], new_argv);
	free(new_argv);
	
	if (!pipe) {
		clog(ERROR, "%s: Cant popen (%s)\n", __func__,
			strerror(errno));
		goto out;
	}
	file_no = fileno(pipe);
	if (modus == MODUS_WRITE) {
		send_buffer(socket_fd, AWAITING_DATA_MSG);
		while ((ret = recv_bin_buffer(socket_fd, buf, sizeof(buf))) > 0)
			write(file_no, buf, ret);
	} else {
		while ((ret = read(file_no, buf, MAXLINE - 1)) > 0) {
			if (send_bin_buffer(socket_fd, buf, ret) < 0) {
				ret = -1;
				goto reread_out;
			}
		}
	}
reread_out:
out:
	if (pipe)
		pclose(pipe);
	return ret;
}

static int com_db_read(int socket_fd, int argc, char **argv) 
{
	return dbtool_cmd_handler(MODUS_READ, socket_fd, argc, argv);
}
static int com_db_write(int socket_fd, int argc, char **argv) 
{
	return dbtool_cmd_handler(MODUS_WRITE, socket_fd, argc, argv);
}

/* server info */
static int com_si(int socket_fd, int argc, char **argv)
{
	char buf[STRINGSIZE];
	char buf2[STRINGSIZE];
	int n;

	uptime_str(server_uptime(GET), buf);
	mmd_lock();
	snprintf(buf2, sizeof(buf2), "up: %s\nplayed: %u\n"
		"connections (total/accepted): %u/%u\n"
		"active connections: %u\n"
		"loglevel: %i\n"
		"supported audio formats: %s\n"
		"%n",
		buf, mmd->num_played, num_connects(GET),
		mmd->num_commands, mmd->active_connections, 
		conf.loglevel_arg, 
		afs_supported_audio_formats(), 
		&n
	);
	mmd_unlock();
	buf2[n] = '\0';
	if (send_buffer(socket_fd, buf2) < 0)
		return -1;
	return 0;
}

/* version */
static int com_version(int socket_fd, int argc, char **argv)
{
	if (send_buffer(socket_fd, "para_server version " VERSION  " | " 
			COPYRIGHT "\n" "built: " BUILD_DATE " (" SYSTEM ", " 
			CC_VERSION ")\n ") < 0)
		return -1;
	return 0;
}

/* sc */
static int com_sc(int socket_fd, int argc, char **argv)
{
	char buf[STRINGSIZE];
	int numplayed, old_numplayed = 0;
repeat:
	mmd_lock();
	numplayed = mmd->num_played;
	if (numplayed != old_numplayed)
		strcpy(buf, mmd->filename);
	mmd_unlock();
	if (numplayed != old_numplayed) {
		old_numplayed = numplayed;
		strcat(buf, "\n");
		if (send_buffer_ll(socket_fd, buf, DEBUG) < 0) 
			return -1;
	}	
	usleep(500000);
	goto repeat;
}

/* sb */
static int com_sb(int socket_fd, int argc, char **argv)
{
	char buf[STRINGSIZE];
	int nr = -1; 	/* status bar will be printed that many 
			 * times. Negative value means: print 
			 * forever
			 */
	if (argc)
		nr = atoi(argv[1]);
	while (nr) {
		mmd_lock();
		compute_sb_string(mmd, buf);
		strcat(buf, "\n");
		mmd_unlock();
		if (send_buffer_ll(socket_fd, buf, DEBUG) < 0)
			return -1;
		if (nr == 1)
			return 0;
		usleep(500000);
		if (nr > 0) 
			nr--;
	}
	return 0;
}

/* stat */
static int com_stat(int socket_fd, int argc, char **argv)
{
	char buf[STRINGSIZE];
	int nr = -1; 	/* status will be printed that many 
			 * times. Negative value means: print 
			 * forever
			 */
	int numplayed = 0; /* used to detect song change */
	struct misc_meta_data nmmd;

	if (argc)
		nr = atoi(argv[1]);
	while (nr) {
 		mmd_dup(&nmmd);
		if (compute_status_volatile(buf, &nmmd) && 
				send_buffer_ll(socket_fd, buf, INFO) < 0)
			return -1;
		if (send_buffer_ll(socket_fd, mmd->dbinfo, INFO) < 0)
			return -1;
		if (numplayed != mmd->num_played) {
			compute_status(buf, sizeof(buf), &nmmd);
			if (send_buffer_ll(socket_fd, buf, INFO) < 0)
				return -1;
			numplayed = mmd->num_played;
		}
		usleep(500000);
		if (nr > 0) 
			nr--;
	}
	return 0;
}

/* help */
static int com_help(int socket_fd, int argc, char **argv)
{
	char buf[STRINGSIZE];
	struct linked_cmd_list* cmd_ptr;
	char perms[MAXLINE];
	int server_cmd;

	if (!argc) { 
		/* no argument given, print list of commands */
		clog(DEBUG, "%s: sending list of commands\n", __func__);
		if (send_buffer(socket_fd, "Name\tDescription\n") < 0)
			return -1;
		return lcl_send_descriptions(socket_fd);
	}
	/* argument given for help */ 
	clog(DEBUG, "%s: sending help for specified command \"%s\"\n", 
		__func__, argv[1]);
	cmd_ptr = lcl_find(argv[1]);
	if (!cmd_ptr)
		return send_buffer(socket_fd, "Unknown command\n");

	server_cmd = (cmd_ptr->cmd->handler != com_db_read 
		&& cmd_ptr->cmd->handler != com_db_write);

	cmd_perms_itohuman(cmd_ptr->cmd->perms, perms);

	snprintf(buf, STRINGSIZE - 1, 
		"NAME:\n~~~~~\n\t%s -- %s\n\n"
		"USAGE:\n~~~~~~\n\t%s\n\n"
		"DESCRIPTION:\n~~~~~~~~~~~~\n%s\n"
		"HANDLER/MODE:\n~~~~~~~~~~~~~\n"
		"This command is handled by %s(%s mode)\n\n"
		"PERMS:\n~~~~~~\n"
		"Needed privileges for %s: %s\n", 
		argv[1],
		cmd_ptr->cmd->description,
		cmd_ptr->cmd->usage,
		cmd_ptr->cmd->help,
		cmd_handler_itohuman(cmd_ptr->cmd->handler),
		cmd_ptr->cmd->modus == MODUS_WRITE?  "write" : "read", 
		argv[1],
		perms
	);
	return send_buffer(socket_fd, buf);
}

/* hup */
static int com_hup(int socket_fd, int argc, char **argv)
{
	kill(getppid(), SIGHUP);
//	printf("%i\n", getppid());
	return 0;
}

/* perms */
static int com_perms(int socket_fd, int argc, char **argv)
{

	if (send_buffer(socket_fd, "Name\tHandler  \tModus\tPermissions\n") < 0)
		return -1;
	return lcl_send_perms(socket_fd);
}

/* pic */
static int com_pic(int socket_fd, int argc, char **argv)
{
	char buf[STRINGSIZE];
	int file_no, n;

	if (dbtool_has_pic) 
		return com_db_read(socket_fd, argc, argv);
	
	file_no = open(PICDIR "/no_pics.jpg", O_RDONLY);
	if (!file_no)
		return -1;
	while ((n = read(file_no, buf, STRINGSIZE - 1))) {
		if (n < 0) {
			close(file_no);
			return -1;			
		}
		if (send_bin_buffer(socket_fd, buf, n) < 0)
			return -1;
	}
	close(file_no);
	return 0;
	
}

/* term */
static int com_term(int socket_fd, int argc, char **argv)
{
//	sigint_sigterm_handler(1);
	kill(getppid(), SIGTERM);
	return 0;
}

/* cs */
static int com_cs(int socket_fd, int argc, char **argv)
{
	char buf[MAXLINE];
	int ret;
	char **stream_list = NULL;

	mmd_lock();
	if (!argc) {
		sprintf(buf, "%s\n", mmd->stream_name);
		ret = send_buffer(socket_fd, buf); 
		goto out;
	}
	if (!strcmp(mmd->stream_name, argv[1])) {
		sprintf(buf, "stream is already %s\n", argv[1]);
		ret = send_buffer(socket_fd, buf);
		goto out;
	}
	if (get_stream_list(&stream_list) <= 0) {
		ret = -1;
		goto out;
	}
	ret = find_stream(argv[1], stream_list);
	if (ret >= 0) {
		clog(INFO, "%s: changing stream to %s\n", __func__, argv[1]);
		strcpy(mmd->stream_name, argv[1]);
		sprintf(buf, "stream changed to %s\n", argv[1]);
		ret = send_buffer(socket_fd, buf);
	} else {
		clog(WARNING, "%s: Request for unknown stream: %s\n", __func__,
			argv[1]);
		ret = -1;
		send_buffer(socket_fd, "No such stream\n");
	}
out:
	mmd_unlock();
	if (stream_list)
		free_stream_list(stream_list);
	return ret;
}

/* streams */
static int com_streams(int socket_fd, int argc, char **argv)
{
	char buf[STRINGSIZE];
	int ret, num_streams;
	char **p, **stream_list = NULL;
	
	mmd_lock();
	num_streams = get_stream_list(&stream_list);
	mmd_unlock();
	if (num_streams <= 0)
		return -1;
	for (p = stream_list; *p; p++) {
		sprintf(buf, "%s\n", *p);
		if (send_buffer(socket_fd, buf) < 0)
			return -1;
	}
	ret = 1;
	free_stream_list(stream_list);
	return ret;
}

/* helper function for ps/ns */
static int cycle_stream(int socket_fd, int argc, char **argv, int previous)
{
	char buf[STRINGSIZE];
	int ret = -1, num_streams, stream_num;
	char **stream_list = NULL;

	mmd_lock();
	num_streams = get_stream_list(&stream_list);
	if (num_streams <= 0)
		goto out;
	stream_num = find_stream(mmd->stream_name, stream_list);
	if (stream_num < 0)
		stream_num = 0;
	else if (previous)
		stream_num = stream_num? stream_num - 1 : num_streams - 1;
	else
		stream_num = (stream_num == num_streams - 1)? 
			0 : stream_num + 1;
	strcpy(mmd->stream_name, stream_list[stream_num]);
	sprintf(buf, "stream changed to %s\n", stream_list[stream_num]);
	ret = send_buffer(socket_fd, buf); 
out:
	mmd_unlock();
	if (stream_list)
		free_stream_list(stream_list);
	return ret;
}

/* ns */
static int com_ns(int socket_fd, int argc, char **argv)
{
	return cycle_stream(socket_fd, argc, argv, 0);
}

/* ps */
static int com_ps(int socket_fd, int argc, char **argv)
{
	return cycle_stream(socket_fd, argc, argv, 1);
}

/* play */
static int com_play(int socket_fd, int argc, char **argv)
{
	int ret = -1;
	char **stream_list = NULL;
	
	mmd_lock();
	if (argc) {
		if (get_stream_list(&stream_list) <= 0)
			goto out;
		ret = find_stream(argv[1], stream_list);
		if (ret < 0) {
			ret = -1;
			send_buffer(socket_fd, "no such stream\n");
			goto out;
		}
		strcpy(mmd->old_stream_name, mmd->stream_name);
		strcpy(mmd->stream_name, stream_list[ret]);
		if (strcmp(mmd->stream_name, mmd->old_stream_name))
			mmd->afs_status_flags |= AFS_NEXT;
	}
	mmd->afs_status_flags |= AFS_PLAYING;
	ret = 1;
out:
	if (stream_list)
		free_stream_list(stream_list);
	mmd_unlock();
	return ret;
}

/* stop */
static int com_stop(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags &= ~AFS_PLAYING;
	mmd->afs_status_flags &= ~AFS_REPOS;
	mmd->afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 0;
}

/* pause */
static int com_pause(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags &= ~AFS_PLAYING;
	mmd->afs_status_flags &= ~AFS_NEXT;
	mmd_unlock();
	return 0;
}

/* next */
static int com_next(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 0;
}

/* nomore */
static int com_nomore(int socket_fd, int argc, char **argv)
{
	mmd_lock();
	mmd->afs_status_flags ^= AFS_NOMORE;
	mmd_unlock();
	return 0;
}
/* ff */
static int com_ff(int socket_fd, int argc, char **argv)
{
	int ret, sign = 1;
	char c;
	unsigned int i, skip;

	if (!argc)
		return -1;
	if (!(ret = sscanf(argv[1], "%u%c", &i, &c)))
		return -1;
	if (ret > 1 && c == '-') 
		sign = -1; /* jmp backwards */
	mmd_lock();
	if (!mmd->size)
		goto out;
	skip = i * ((mmd->chunks_total + 5) / 10)  / 
		((mmd->seconds_total + 5) / 10);
	if (sign == 1 && skip + mmd->chunks_sent > mmd->chunks_total) {
		mmd->afs_status_flags |= AFS_NEXT;
		goto out;
	}
	mmd->afs_status_flags |= AFS_REPOS;
	mmd->afs_status_flags &= ~AFS_NEXT;
	if (sign == -1 && mmd->chunks_sent < skip) {
		mmd->chunks_sent = 0;
		goto out;
	}
	mmd->chunks_sent += skip * sign;
out:
	mmd_unlock();
	return 0;
}

/* jmp */
static int com_jmp(int socket_fd, int argc, char **argv)
{
	long unsigned int i;

	if (!argc)
		return -1;
	if (sscanf(argv[1], "%lu", &i) <= 0)
		return -1;
	if (i > 100) 
		i = 100;
	mmd_lock();
	clog(DEBUG, "%s: jumping to %lu/%lu\n", __func__, 
		mmd->chunks_total / 100 * i, mmd->chunks_total);
	mmd->chunks_sent = mmd->chunks_total / 100 * i;
	mmd->afs_status_flags |= AFS_REPOS;
	mmd->afs_status_flags &= ~AFS_NEXT;
	mmd_unlock();
	return 0;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms. 
 * Returns 0 if perms are sufficient, -1 otherwise.
 */
int check_perms(unsigned int perms, struct linked_cmd_list *cmd_ptr) 
{
	clog(DEBUG, "%s: checking permissions\n", __func__);
	return (cmd_ptr->cmd->perms & perms) < cmd_ptr->cmd->perms ? -1 : 0;
}

/*
 * Parse first string from *cmd and lookup in table of valid commands.
 * On error, NULL is returned. 
 */
struct linked_cmd_list *parse_cmd(char *cmdstr) 
{
	char buf[MAXLINE];
	struct linked_cmd_list *cmd_ptr;
	int n = 0;

	sscanf(cmdstr, "%200s%n", buf, &n);
	if (!n)
	        return NULL;
	buf[n] = '\0';
	cmd_ptr = lcl_find(buf);
	if (cmd_ptr)
		clog(DEBUG, "%s: Found command %s. Perms: %u\n", __func__,
			cmd_ptr->cmd->name, cmd_ptr->cmd->perms);
	else 
		clog(ERROR, "%s: Unknown Command: %s\n", __func__, buf);
	return cmd_ptr;
}
